// Fichier Registry.cpp

#include <string>
#include <objbase.h>
#include <assert.h>
#include "registry.h"
#include "ConversionChaines.h"

struct REG_DATA
{
    const char* pszKey;
    const char* pszValue;
    const char* pszData;
};

// Dfinit la cl donne et sa valeur.
bool FixerCleEtValeur (const std::string & Cle, const char* SousCle, const char* Valeur);

// Ouvre une cl et dfinit une valeur.
bool FixerValeurDansCle (const std::string & Cle, const std::string & ValeurNommee, const char* Valeur);

// Convertit un CLSID en une chaine de caractres.
std::string CLSIDtochar(REFCLSID clsid);

// Supprime szKeyChild et tous ses descendants.
LONG recursiveDeleteKey(HKEY hKeyParent, const std::string & CleEnfant);

HRESULT UnregisterServerEx(const REG_DATA regData[]);


std::string ObtenirNomCompletModule (HMODULE hModule)
{
   // Taille d'un CLSID en tant que chaine.
   const int TAILLE_MAX_NOMFICHIER = 512;
	char szModule[TAILLE_MAX_NOMFICHIER];
	DWORD dwResult = GetModuleFileName(hModule, szModule, sizeof(szModule)/sizeof(char));
	assert(dwResult);
   return szModule;
}

std::string ContruireCLSIDBrut (REFCLSID clsid)
{
	// Convertit le CLSID en un char.
   return CLSIDtochar(clsid);
}

std::string ContruireCLSID (const std::string & clsid_brut)
{
	// construit la cl CLSID\\{...}
   return std::string ("CLSID\\") + clsid_brut;
}

bool EnregistrementCle (const std::string & Cle,
                        const std::string & NomConvivial)
{
	return FixerCleEtValeur (Cle, 0, NomConvivial.c_str ());
}

bool EnregistrementProgID (const std::string & Cle,
                           const std::string & ProgID)
{
   const std::string SOUSCLE_PROGID = "ProgID";
	return FixerCleEtValeur (Cle, SOUSCLE_PROGID.c_str (), ProgID.c_str ());
}

bool EnregistrementTypeServeur (const std::string & NomModule,
                                const std::string & NomModuleComplet,
                                const std::string & Cle,
                                const char * ThreadingModel)
{
   const std::string EXTENSION_SERVEUR_EXTERNE = ".exe";
	// Ajoute la sous-cl "nom de fichier du serveur" sous la cl CLSID.
   bool ResultatEnregistrement;
   if(NomModule.find (EXTENSION_SERVEUR_EXTERNE) == std::string::npos)
	{
      const std::string NOMCLE_SERVEUR_INTERNE = "InprocServer32";
      std::string CleInterne = Cle + "\\" + NOMCLE_SERVEUR_INTERNE;
		ResultatEnregistrement =
         FixerCleEtValeur (Cle,
                           NOMCLE_SERVEUR_INTERNE.c_str (),
                           NomModuleComplet.c_str ()) &&
		   FixerValeurDansCle (CleInterne.c_str (),
                             "ThreadingModel",
                             ThreadingModel);
	}
	else
   {
      const std::string NOMCLE_SERVEUR_EXTERNE = "LocalServer32";
		ResultatEnregistrement =
         FixerCleEtValeur (Cle,
                           NOMCLE_SERVEUR_EXTERNE.c_str (),
                           NomModuleComplet.c_str ());
   }
   return ResultatEnregistrement;
}

bool EnregistrementVersionDependent (const std::string & ProgID,
                                     const std::string & NomConvivial,
                                     const std::string & clsid_brut)
{
	return FixerCleEtValeur (ProgID, 0, NomConvivial.c_str ()) &&
	       FixerCleEtValeur (ProgID, "CLSID", clsid_brut.c_str ());
}

bool EnregistrementCleVersionIndependent (const std::string & Cle,
                                          const std::string & NomIndependantDeVersion)
{
   const std::string SOUSCLE_VERSIONINDEPENDENTPROGID = "VersionIndependentProgID";
   return FixerCleEtValeur (Cle,
                            SOUSCLE_VERSIONINDEPENDENTPROGID.c_str (),
                            NomIndependantDeVersion.c_str ());
}

bool EnregistrementVersionIndependent (const std::string & ProgID,
                                       const std::string & NomConvivial,
                                       const std::string & clsid_brut,
                                       const std::string & NomIndependantDeVersion)
{
   return FixerCleEtValeur (NomIndependantDeVersion,
                            0,
                            NomConvivial.c_str ()) && 
	       FixerCleEtValeur (NomIndependantDeVersion,
                            "CLSID",
                            clsid_brut.c_str ()) &&
	       FixerCleEtValeur (NomIndependantDeVersion,
                            "CurVer",
                            ProgID.c_str ());
}

// Inscrit le composant dans le registre
HRESULT RegisterServer(const std::string & NomModule, // nom du module, qu'il soit DLL ou EXE
                       REFCLSID clsid,               // ID de classe
                       const std::string & NomConvivial,   // nom convivial
                       const std::string & NomIndependantDeVersion,   // Programmatique
                       const std::string & ProgID,         // ID
					        const char * ThreadingModel) // modle multithread ou non
{
	// obtient l'emplacement du serveur
   std::string NomModuleComplet = ObtenirNomCompletModule (GetModuleHandle(NomModule.c_str ()));

   std::string clsid_brut = ContruireCLSIDBrut (clsid);
   std::string Cle = ContruireCLSID (clsid_brut);

   // Ajoute le CLSID au registre.
   HRESULT hr = S_OK;
   if (!EnregistrementCle (Cle, NomConvivial) ||
       !EnregistrementTypeServeur (NomModule, NomModuleComplet, Cle, ThreadingModel) ||
	    !EnregistrementProgID (Cle, ProgID) ||
       !EnregistrementCleVersionIndependent (Cle, NomIndependantDeVersion) ||
       !EnregistrementVersionIndependent    (ProgID, NomConvivial, clsid_brut, NomIndependantDeVersion) ||
       !EnregistrementVersionDependent      (ProgID, NomConvivial, clsid_brut))
      hr = S_FALSE;

	return hr;
}

// Supprime le composant du registre.
LONG UnregisterServer(REFCLSID clsid,             // ID de classe
                      const std::string & NomIndependantDeVersion, // Programmatique
                      const std::string & ProgID)       // ID
{
   std::string clsid_brut = ContruireCLSIDBrut (clsid);
   std::string Cle = ContruireCLSID (clsid_brut);

	// supprime la cl CLSID\{...}
	LONG lResult = recursiveDeleteKey(HKEY_CLASSES_ROOT, Cle);
	assert((lResult == ERROR_SUCCESS) || (lResult == ERROR_FILE_NOT_FOUND)); // Subkey may not exist.

	// supprime la cl version-independent ProgID.
	lResult = recursiveDeleteKey(HKEY_CLASSES_ROOT, NomIndependantDeVersion);
	assert((lResult == ERROR_SUCCESS) || (lResult == ERROR_FILE_NOT_FOUND)); // Subkey may not exist.

	// supprime la cl ProgID.
	lResult = recursiveDeleteKey(HKEY_CLASSES_ROOT, ProgID);
	assert((lResult == ERROR_SUCCESS) || (lResult == ERROR_FILE_NOT_FOUND)); // Subkey may not exist.

	return S_OK;
}

// Convertit un CLSID en une chaine.
std::string CLSIDtochar(REFCLSID clsid)
{
	// Get CLSID
	LPOLESTR wszCLSID = 0;
	HRESULT hr = StringFromCLSID(clsid, &wszCLSID);
	assert(SUCCEEDED(hr));
   
   std::string clsid_obtenu = WStringToString (wszCLSID);

	// Libre la mmoire.
	CoTaskMemFree(wszCLSID);

   return clsid_obtenu;
}

// supprime une cl et tous ses descendants.
LONG recursiveDeleteKey(HKEY hKeyParent,           // Parent de la cl  supprimer
                        const std::string & CleEnfant)  // cl  supprimer
{
	// Ouvre l'enfant.
	HKEY hKeyEnfant;
	LONG lRes = RegOpenKeyEx(hKeyParent, CleEnfant.c_str (), 0, KEY_ALL_ACCESS, &hKeyEnfant);
	if(lRes != ERROR_SUCCESS)
		return lRes;

	// Enumre tous les decendents de cet enfant.
	FILETIME time;
   const int TAILLE_TAMPON = 256;
	char szBuffer[TAILLE_TAMPON];
	DWORD dwSize = TAILLE_TAMPON;
	while(RegEnumKeyEx(hKeyEnfant, 0, szBuffer, &dwSize, 0, 0, 0, &time) == S_OK)
	{
		// supprime les decendants de cet enfant.
		lRes = recursiveDeleteKey(hKeyEnfant, szBuffer);
		if(lRes != ERROR_SUCCESS)
		{
			// nettoyage avant fin.
			RegCloseKey(hKeyEnfant);
			return lRes;
		}
		dwSize = TAILLE_TAMPON;
	}

	// Ferme l'enfant.
	RegCloseKey(hKeyEnfant);

	// Supprime cet enfant.
	return RegDeleteKey(hKeyParent, CleEnfant.c_str ());
}

// Cre une cl et dfinit sa valeur.
bool FixerCleEtValeur(const std::string & Cle, const char* SousCle, const char* Valeur)
{
   std::string KeyBuf (Cle);

	// Ajoute nom de sous-cl au tampon.
	if(SousCle)
	{
		KeyBuf += "\\";
		KeyBuf += SousCle;
	}

	// Cre et ouvre cl plus sous-cl.
	HKEY hKey;
	long lResult = RegCreateKeyEx(HKEY_CLASSES_ROOT, KeyBuf.c_str (), 0, 0, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, 0, &hKey, 0);
	if(lResult != ERROR_SUCCESS)
		return false;

	// dfinit la valeur.
	if(Valeur)
		RegSetValueEx(hKey, NULL, 0, REG_SZ, (BYTE *)Valeur,
                    static_cast<DWORD> (strlen(Valeur)+1));

	RegCloseKey(hKey);
	return true;
}

// Ouvre une cl et dfinit une valeur.
bool FixerValeurDansCle (const std::string & Cle, const std::string & ValeurNommee, const char* Valeur)
{
   std::string KeyBuf (Cle);

	// Cre et ouvre cl plus sous-cl.
	HKEY hKey;
	long lResult = RegOpenKeyEx(HKEY_CLASSES_ROOT, KeyBuf.c_str (), 0, KEY_SET_VALUE, &hKey);
	if(lResult != ERROR_SUCCESS)
		return false;

    // dfinit la valeur.
	if(Valeur)
		RegSetValueEx(hKey, ValeurNommee.c_str (), 0, REG_SZ, (BYTE*)Valeur,
                    static_cast<DWORD>(strlen(Valeur)+1));

	RegCloseKey(hKey);
	return true;
}

HRESULT RegisterServerEx(const REG_DATA regData[], const char* szModuleName)
{
	for(int count = 0;; count++)
	{
		const char* pszKey = regData[count].pszKey;
		const char* pszValue = regData[count].pszValue;
		const char* pszData = regData[count].pszData;

		if(pszKey == 0 && pszValue == 0 && pszData == 0)
			break;

		if(pszData == (const char*)-1)
			pszData = szModuleName;

		HKEY hKey;
		long err = RegCreateKey(HKEY_CLASSES_ROOT, pszKey, &hKey);
		if(err == ERROR_SUCCESS)
		{
			err = RegSetValueEx(hKey, pszValue, 0, REG_SZ, (const BYTE*)pszData,
                             static_cast<DWORD>(strlen(pszData) + 1));
			RegCloseKey(hKey);
		}
		if(err != ERROR_SUCCESS)
		{
			UnregisterServerEx(regData);
			return REGDB_E_WRITEREGDB;
		}
	}
	return S_OK;
}

HRESULT UnregisterServerEx(const REG_DATA regData[])
{
	HRESULT hr = S_OK;

	for(int nEntries = 0;; nEntries++)
		if(regData[nEntries].pszKey == 0 && regData[nEntries].pszValue == 0 && regData[nEntries].pszData == 0)
			break;

	for(int count = nEntries - 1; count >= 0; count--)
		if(RegDeleteKeyA(HKEY_CLASSES_ROOT, regData[count].pszKey) != ERROR_SUCCESS)
			hr = S_FALSE;

	return hr;
}